/*	Capture a frame of video from a V4L2 driver and send it
 *	to stdout. Use vctrl to set image width, height, depth.
 *
 *	This program was written by Bill Dirks.
 *	This program is in the public domain.
 *
 *	gcc -o xcaptest -L/usr/X11R6/lib/ -lXt -lXaw -Wall xcaptest.c
 *
 *	./vcat [-rgb | -bgr] [device-node]
 */


#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/mman.h>
#include <errno.h>

/* Video for Linux Two */
#include <linux/fs.h>
#include <videodev.h>  // <linux/videodev.h>

#define MY_FMT_FIELD  V4L2_FIELD_INTERLACED
#define MY_WIDTH  320
#define MY_HEIGHT 240
// #define MY_PIXEL_FORMAT V4L2_PIX_FMT_BGR24
#define MY_PIXEL_FORMAT V4L2_PIX_FMT_BGR32

int
main(int argc, char *argv[])
{
	char			*device	= NULL;
	int			vid;
	int			err;
	int			i;
	int			n;
	struct v4l2_capability	cap;
        struct v4l2_format	fmt;
	int			bgr	= 0;
	int			rgb	= 0;
	char			*data;
	int                     width;
	int                     height;


	for (i = 1; i < argc; ++i)
	{
		if (strcmp(argv[i], "-rgb") == 0)
			rgb = 1;
		else if (strcmp(argv[i], "-bgr") == 0)
			bgr = 1;
		else if (argv[i][0] == '-')
			continue;
		else
			device = argv[i];
	}
	if (device == NULL)
		device = "/dev/video0";


	vid = open(device, O_RDONLY);
	if (vid < 0)
	{
		fprintf(stderr, "Can't open %s\n", device);
		return 1;
	}

	err = ioctl(vid, VIDIOC_QUERYCAP, &cap);
	if (err)
	{
		fprintf(stderr, "QUERYCAP returned error %d\n", errno);
	        close(vid);
		return 1;
	}
	/*
	if (cap.type != V4L2_TYPE_CAPTURE)
	{
		fprintf(stderr, "Device %s is not a video capture device.\n",
			device);
	        close(vid);
		return 1;
	}
	if (!(cap.flags & V4L2_FLAG_READ))
	{
		fprintf(stderr, "Device %s doesn't support read().\n", device);
	        close(vid);
		return 1;
	}
	*/


	// fmt.type = V4L2_BUF_TYPE_CAPTURE;
	fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	err = ioctl(vid, VIDIOC_G_FMT, &fmt);
	if (err)
	{
		fprintf(stderr, "G_FMT returned error %d\n", errno);
	        close(vid);
		return 1;
	}
        // pixelformat is the composition of four chars 
	//     A<<0 | B<<8 | C<<16 | D<<24
	// see include/linux/videodev.h
	fprintf(stderr, "G_FMT: pixelformat %x Width %d Height %d\n",
	  fmt.fmt.pix.pixelformat, fmt.fmt.pix.width, fmt.fmt.pix.height);

        // fmt.fmt.pix.flags = FMT_FLAG;
	fmt.fmt.pix.field  = MY_FMT_FIELD;
        fmt.fmt.pix.width  = MY_WIDTH;
        fmt.fmt.pix.height = MY_HEIGHT;
	fmt.fmt.pix.pixelformat = MY_PIXEL_FORMAT;
	err = ioctl(vid, VIDIOC_S_FMT, &fmt);
	if (err)
	{
		fprintf(stderr, "S_FMT returned error %d\n", errno);
	        close(vid);
		return 1;
	}

	fprintf(stderr, "image width  %d\n", fmt.fmt.pix.width );
	fprintf(stderr, "image height %d\n", fmt.fmt.pix.height );
	width  = fmt.fmt.pix.width;
	height = fmt.fmt.pix.height;

	data = malloc(fmt.fmt.pix.sizeimage);
	if (data == NULL)
	{
		fprintf(stderr, "malloc(%d) failed\n", fmt.fmt.pix.sizeimage);
	        close(vid);
		return 1;
	}

	n = read(vid, data, fmt.fmt.pix.sizeimage);
	if (n < 0)
	{
		fprintf(stderr, "read() returned error %d\n", errno);
	        close(vid);
		return 1;
	}
	fprintf(stderr, "read returned %d: pixels %d = %d (%dx%d)\n", 
	  n, n/4, width*height, width, height );

	if (n) {
	  for (i=0; i<n/4; i++) 
	    fwrite(data+i*4, (size_t)3, 1, stdout);
	}

	close(vid);

	return 0;
}
