/*      Video Stream Capture Testing Tool

 *      For testing a Video for Linux Two video capture driver without display
 *
 *      This program was wadapted by Olivier Carmona from Bill Dirks source.
 *      This program is in the public domain.
 *
 *      gcc -o vstreamtest -L/usr/X11R6/lib/ -lXt -lXaw -Wall vstreamtest.c
 *
 *      ./vstreamtest [-b] [device-node]
 *
 *      -b      The X Window screen is 32 bits-per-pixel.
 */

/*  This program does not properly detect when the X Windows desktop is
 *  32 bits per pixel. (Some X functions use a value of 24 for the depth
 *  when it is really 32 for reasons I don't understand.) As a workaround,
 *  if you are running a 32-bpp X Windows screen, use -b on the command
 *  line, or uncomment this line:  */
#define X_DEPTH  32


/*  Set these according to your capture device set-up */
#define MY_DEVICE "/dev/video0"
#define MY_STD	  "NTSC"	/* "SECAM", "PAL" or "NTSC" */
#define MY_WIDTH  320
#define MY_HEIGHT 240
#define MY_INPUT 0
#define STREAMBUFS	4
#define FMT_FLAGS V4L2_FIELD_INTERLACED
#define MY_PIXELFORMAT V4L2_PIX_FMT_BGR32
/*  MY_PIXELFORMAT must be defined */
/* V4L2_PIX_FMT_BGR24 */
/* V4L2_PIX_FMT_YUYV */
/* V4L2_PIX_FMT_GREY */
/* V4L2_PIX_FMT_YUV420 */
/* V4L2_PIX_FMT_RGB555 */
/* V4L2_PIX_FMT_RGB565 */
/* V4L2_PIX_FMT_BGR24 */

/*  Uncomment the following to test stuff  */
#define TEST_INPUTS		/* enumerate the inputs */
#define TEST_FORMATS		/* enumerate the image formats */
#define TEST_CONTROLS		/* enumerate the controls */
#define TEST_CAPTURE		/* test video capture */
#define TEST_STREAMING		
#define TEST_SELECT		



#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/mman.h>
#include <errno.h>
#define max(a,b) ((a) > (b) ? (a) : (b))

/* These are needed to use the Videum driver */
#include <linux/fs.h>
#include <linux/kernel.h>
#include <videodev.h>	/* Video for Linux Two */
#include <meteor_v4l2.h>

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Porthole.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>

typedef struct tag_vimage
  {
    struct v4l2_buffer vidbuf;
    char *data;
    XImage *ximage;
  }
VIMAGE;

VIMAGE vimage[STREAMBUFS];
int vid;
int x_depth;
struct v4l2_format fmt;

struct
  {
    int width, height;
  }
capturesize[] =
{
  { 160, 120 } ,
  { 192, 144 } ,
  { 320, 240 } ,
  { 384, 288 } ,
  { 640, 480 } ,
  { 768, 576 } ,
};



void 
stop_capturing ()
{
  int i;

  if (vid >= 0)
    {
      i = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      ioctl (vid, VIDIOC_STREAMOFF, &i);
      for (i = 0; i < STREAMBUFS; ++i)
	{
	  if (vimage[i].data)
	    munmap (vimage[i].data,
		    vimage[i].vidbuf.length);
	  vimage[i].data = NULL;
	  if (vimage[i].ximage)
// memory leak, but XDestroyImage always segfaults...
	    //                              XDestroyImage(vimage[i].ximage);
	    vimage[i].ximage = NULL;
	}
    }
}

int 
start_capturing ()
{
  struct v4l2_requestbuffers req;
  int err;
  int i;

  fmt.fmt.pix.pixelformat = MY_PIXELFORMAT;
  err = ioctl (vid, VIDIOC_S_FMT, &fmt);
  if (err)
    {
      printf ("S_FMT returned error %d\n", errno);
      return 1;
    }
  err = ioctl (vid, VIDIOC_S_FMT, &fmt);
  if (err)
    {
      printf ("S_FMT returned error %d\n", errno);
      return 1;
    }

  req.count = STREAMBUFS;
  req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  err = ioctl (vid, VIDIOC_REQBUFS, &req);
  if (err < 0 || req.count < 1)
    {
      printf ("REQBUFS returned error %d, count %d\n",
	      errno, req.count);
      return 1;
    }
  for (i = 0; i < req.count; ++i)
    {
      vimage[i].vidbuf.index = i;
      vimage[i].vidbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      err = ioctl (vid, VIDIOC_QUERYBUF, &vimage[i].vidbuf);
      if (err < 0)
	{
	  printf ("QUERYBUF returned error %d\n", errno);
	  return 1;
	}
      vimage[i].data = mmap (0, vimage[i].vidbuf.length, PROT_READ,
			     MAP_SHARED, vid,
			     vimage[i].vidbuf.m.offset);
      if ((int) vimage[i].data == -1)
	{
	  printf ("mmap() returned error %d\n", errno);
	  return 1;
	}
    }

  for (i = 0; i < req.count; ++i)
    if ((err = ioctl (vid, VIDIOC_QBUF, &vimage[i].vidbuf)))
      {
	printf ("QBUF returned error %d\n", errno);
	return 1;
      }
  err = ioctl (vid, VIDIOC_STREAMON, &vimage[0].vidbuf.type);
  if (err)
    printf ("STREAMON returned error %d\n", errno);

  return 0;
}

/* --- */
/* WARNING: these must agree with enum v4l2_ctrl_type in videodev2.h */
char * ctrl_type[] = {
		"(unknown)",
		"(int)",
		"(bool)",
		"(menu)",
		"(button)"
};

void 
printctrl (struct v4l2_queryctrl *qc)
{
  printf ("  %-12s %-6s %8d %8d %7d %8d\n",
	  qc->name, ctrl_type[qc->type],
	  qc->minimum, qc->maximum, qc->step, qc->default_value );
}

int
main (int argc, char *argv[])
{
  char my_device[64];
  int err;
  int i;
  int frame_nr = 100;
  struct v4l2_capability cap;
  struct v4l2_streamparm parm;
#ifdef MY_STD
  struct v4l2_standard std;
#endif
#ifdef MY_INPUT
  int input;
#endif
#ifdef TEST_STREAMING
  fd_set rdset;
  struct timeval timeout;
  int n;
#endif
#ifdef TEST_CONTROLS
  struct v4l2_queryctrl qc;
#endif
  struct v4l2_buffer tempbuf;
  int argi;
  int delta;
  double timeperframe;
  struct timeval start_time, cur_time;

#ifdef CONFIG_V4L2_PERFORMANCE
  struct v4l2_performance perf;
#endif


/*->	Put in the device node name */
  strcpy (my_device, MY_DEVICE);
  for (argi = 1; argi < argc; ++argi)
    {
      if (strcmp (argv[argi], "-b") == 0)
	  {
	    x_depth = 32;
	    continue;
	  }
	  if (strcmp (argv[argi], "-f") == 0)
	  {
	    argi ++;
		frame_nr = atoi( argv[argi] );
		continue;
	  }
      strcpy (my_device, argv[argi]);
    }

  vid = open (my_device, O_RDONLY);
  if (vid < 0)
    {
      printf ("No video device \"%s\"\n", my_device);
      return 1;
    }

  err = ioctl (vid, VIDIOC_QUERYCAP, &cap);
  if (err)
    {
      printf ("QUERYCAP returned error %d\n", errno);
      return 1;
    }
  printf ("Device node:  %s\n", my_device);
  printf ("Device:       %s\n", cap.driver);
  printf ("Capabilities:");
  if (cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)
    printf (" video capture;");
  if (cap.capabilities & V4L2_CAP_VBI_CAPTURE)
    printf (" VBI capture;");
  if (cap.capabilities & V4L2_CAP_RDS_CAPTURE)
    printf (" RDS capture;");
  if (cap.capabilities & V4L2_CAP_READWRITE)
    printf (" capture via read();");
  if (cap.capabilities & V4L2_CAP_STREAMING)
    printf (" streaming capture;");
  if (cap.capabilities & V4L2_CAP_ASYNCIO)
    printf (" asynchronous I/O;");
  if (cap.capabilities & V4L2_CAP_AUDIO)
    printf (" audio;");
  if (cap.capabilities & V4L2_CAP_TUNER)
    printf (" tuner;");
  if (cap.capabilities & V4L2_CAP_RADIO)
    printf (" radio;");
  printf ("\n");

  if ( !(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE) )
    {
      printf ("Not a capture device.\n");
      return 1;
    }

#ifdef TEST_INPUTS
  printf ("Video inputs on this device:\n");
  for (i = 0, err = 0; err == 0; ++i)
    {
      struct v4l2_input inp;
      inp.index = i;
      err = ioctl (vid, VIDIOC_ENUMINPUT, &inp);
      if (!err)
	printf ("  %d: \"%s\" is a%s, audio %x tuner %x\n", i,
		inp.name,
		(inp.type == V4L2_INPUT_TYPE_TUNER) ?
		" TV tuner" : "n analog input",
		inp.audioset, 
		inp.tuner
	  );
    }
#ifdef MY_INPUT
  input = MY_INPUT;
  err = ioctl (vid, VIDIOC_S_INPUT, &input);
  if (err)
    printf ("S_INPUT returned error %d\n", errno);
#endif
#endif

#ifdef MY_STD
  ioctl (vid, VIDIOC_G_STD, &std);
#endif
  printf ("Video standards supported:");
  for (i = 0, err = 0; err == 0; ++i)
    {
      struct v4l2_enumstd estd;
      estd.index = i;
      err = ioctl (vid, VIDIOC_ENUMSTD, &estd);
      if (!err)
	{
	  printf ("%c %s", i ? ',' : ' ', estd.std.name);
#ifdef MY_STD
	  if (strcmp (MY_STD, estd.std.name) == 0)
	    std = estd.std;
#endif
	}
    }
  printf ("\n");

#if 0 // def TEST_FORMATS
  printf ("Video capture image formats supported:\n");
  for (i = 0, err = 0; err == 0; ++i)
    {
      struct v4l2_fmtdesc fmtd;
      fmtd.index = i;
      err = ioctl (vid, VIDIOC_ENUM_PIXFMT, &fmtd);
      if (!err)
	printf ("  %d: %s  (%s)\n", i,
		fmtd.description,
		(fmtd.flags & V4L2_FMT_FLAG_COMPRESSED) ?
		"compressed" : "uncompressed");
    }
#endif

#ifdef TEST_CONTROLS
  printf ("Controls supported:\n");
  printf ("  Label        type    Minimum  Maximum    Step  Default"
	  "  Category\n");
  qc.id = V4L2_CID_BRIGHTNESS;
  if (ioctl (vid, VIDIOC_QUERYCTRL, &qc) == 0)
    printctrl (&qc);
  qc.id = V4L2_CID_CONTRAST;
  if (ioctl (vid, VIDIOC_QUERYCTRL, &qc) == 0)
    printctrl (&qc);
  qc.id = V4L2_CID_SATURATION;
  if (ioctl (vid, VIDIOC_QUERYCTRL, &qc) == 0)
    printctrl (&qc);
  qc.id = V4L2_CID_HUE;
  if (ioctl (vid, VIDIOC_QUERYCTRL, &qc) == 0)
    printctrl (&qc);
#endif

#ifndef TEST_CAPTURE
  return 0;
#endif

  if (!(cap.capabilities & V4L2_CAP_STREAMING))
    {
      printf ("Device does not support streaming capture.\n");
      return 1;
    }

  parm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  ioctl (vid, VIDIOC_G_PARM, &parm);
#ifdef TEST_FRAMERATE
  parm.parm.capture.capturemode |= V4L2_CAP_TIMEPERFRAME;
  parm.parm.capture.timeperframe = TEST_FRAMERATE;
#endif
  err = ioctl (vid, VIDIOC_S_PARM, &parm);
  if (err)
    printf ("S_PARM returned error %d\n", errno);

#ifdef MY_STD
  err = ioctl (vid, VIDIOC_S_STD, &std);
  if (err)
    printf ("S_STD returned error %d\n", errno);
#endif

  fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  ioctl (vid, VIDIOC_G_FMT, &fmt);
  fmt.fmt.pix.field |= FMT_FLAGS;
  fmt.fmt.pix.width  = MY_WIDTH;
  fmt.fmt.pix.height = MY_HEIGHT;
  ioctl (vid, VIDIOC_S_FMT, &fmt);

  if (start_capturing ())
    {
      return 1;
    }

  printf ("Capturing %dx%d \"%4.4s\" images\n",
	  fmt.fmt.pix.width,
	  fmt.fmt.pix.height,
	  (char *) &(fmt.fmt.pix.pixelformat));
  printf ("Images are %d bytes each\n", fmt.fmt.pix.sizeimage);

  gettimeofday( &start_time, NULL );

  for (i=0; i<frame_nr; i++)
    {
      FD_ZERO (&rdset);
      FD_SET (vid, &rdset);
      timeout.tv_sec = 1;
      timeout.tv_usec = 0;
      n = select (vid + 1, &rdset, NULL, NULL, &timeout);
      err = -1;
      if (n == -1)
    	fprintf (stderr, "select error.\n");
      else if (n == 0)
	    fprintf (stderr, "select timeout\n");
      else if (FD_ISSET (vid, &rdset))
	    err = 0;

#ifdef CONFIG_V4L2_PERFORMANCE
      if (ioctl (vid, VIDIOC_G_PERF, &perf) != 0)
	    printf ("G_PERF returned error code %d\n", errno);
      else
        printf("Frames %d\n",perf.frames);
#else
	  gettimeofday( &cur_time, NULL );
	  delta = (cur_time.tv_sec - start_time.tv_sec)*1000
			+ (cur_time.tv_usec - start_time.tv_usec)/1000;
	  fprintf(stderr, "%6d: frame %d \r", delta, i );
#endif

      if (err == 0)
	  {
	    tempbuf.type = vimage[0].vidbuf.type;
	    err = ioctl (vid, VIDIOC_DQBUF, &tempbuf);
	    if (err)
	    {
	      perror ("DQBUF");
	      // Needed if not tempbuf not init correctly
	      continue;
	    }
	    err = ioctl (vid, VIDIOC_QBUF, &tempbuf);
	    if (err)
	      perror ("QBUF");
	  }
    }
  timeperframe = (double)delta / (double)frame_nr;
  printf("\nAverage time per frame: %6.3f msec\n",
	timeperframe );
  return 0;
}
